package org.linotte.moteur.xml.alize.kernel;

import java.math.BigDecimal;

import javax.swing.SwingUtilities;
import javax.swing.text.BadLocationException;

import org.alize.kernel.AKContextI;
import org.alize.kernel.AKException;
import org.alize.kernel.AKJob;
import org.alize.kernel.AKJobContext;
import org.linotte.frame.moteur.FrameProcess;
import org.linotte.moteur.entites.Acteur;
import org.linotte.moteur.entites.Livre;
import org.linotte.moteur.entites.Role;
import org.linotte.moteur.exception.ErreurException;
import org.linotte.moteur.exception.FonctionException;
import org.linotte.moteur.exception.LectureException;
import org.linotte.moteur.exception.LinotteException;
import org.linotte.moteur.exception.MathematiquesException;
import org.linotte.moteur.exception.Messages;
import org.linotte.moteur.exception.PositionException;
import org.linotte.moteur.exception.RetournerException;
import org.linotte.moteur.exception.StopException;
import org.linotte.moteur.outils.Chaine;
import org.linotte.moteur.xml.Version;
import org.linotte.moteur.xml.alize.kernel.processus.Processus;
import org.linotte.moteur.xml.appels.CalqueParagraphe;
import org.linotte.moteur.xml.appels.Condition.ETAT_CONDITION;
import org.linotte.moteur.xml.appels.Ring;

public class Job extends AKJob {

	private Processus currentProcessus;

	private Job pere;

	public Job(AKJobContext pcontext) {
		super(pcontext);
	}

	private boolean debogueur = false;

	private Exception erreurAPropager;

	// Indique que ce job est execute dans un thread (verbe appeler)
	private boolean parallelise = false;

	@Override
	public void execute(AKContextI context) throws Exception {
		try {
			currentProcessus = (Processus) getFirstProcessus();
			boolean boucle = true;
			while (boucle)
				try {
					boucle = false;
					while ((currentProcessus = (Processus) currentProcessus.execute(this)) != null) {
						if (erreurAPropager != null) {
							throw erreurAPropager;
						}
						// Souffleur ?
						if (currentProcessus.getSouffleurs() != null) {
							Processus temp = currentProcessus;
							for (Processus souffleur : currentProcessus.getSouffleurs()) {
								currentProcessus = souffleur;
								while (currentProcessus != null)
									currentProcessus = (Processus) currentProcessus.execute(this);
							}
							currentProcessus = temp;
						}

						if (!isRunning() && erreurAPropager == null) {
							throw new StopException();
						}
						// Mode debogue ?
						if (debogueur) {
							afficheDeboggueur();
						}
						
						// Pour partager le temps de traitement entre les jobs
						if (isParallelise())
							Thread.yield();
					}

				} catch (ErreurException e) {
					if (!isDead()) {
						fillStack(e);
						/**
						 * s'il n'est pas mort, on propage l'erreur, sinon, on
						 * ne fait rien (bug sur les fonctions).
						 */
						boucle = handlerError(e);
					}
				} catch (LectureException e) {
					fillStack(e);
					// Cas remonté par Pat dans l'import d'un livre :
					// http://langagelinotte.free.fr/punbb/viewtopic.php?id=273
					if (e.getCause() != null) {
						boucle = handlerError(e.getCause());
					} else
						throw e;
				}

		} catch (RetournerException e) {
			throw e;
		} catch (StopException e) {
			throw e;
		} catch (LectureException e) {
			throw e;
		} catch (StackOverflowError e) {
			throw new AKException("trop de boucles ou d'appels imbriqués ! Le livre doit s'arrêter...");
		} catch (FonctionException e) {
			throw new LectureException(e.getCauseLectureException().getException(), e.getCauseLectureException().getPosition());
		} catch (LinotteException e) {
			throw new LectureException(e, currentProcessus.getPosition());
		} catch (IndexOutOfBoundsException e) {
			if (Version.isBeta())
				e.printStackTrace();
			throw new LectureException(new PositionException(), currentProcessus.getPosition());
		} catch (ArithmeticException e) {
			throw new LectureException(new MathematiquesException(), currentProcessus.getPosition());
		} finally {
			if (erreurAPropager != null) {
				throw erreurAPropager;
			}
		}
	}

	private void fillStack(LinotteException e) {
		Livre livre = ((JobContext) getContext()).getLivre();
		if (livre != null) {
			KernelStack kernelStack = livre.getKernelStack();
			e.setKernelStack((KernelStack) kernelStack.clone());
		}
	}

	private boolean handlerError(LinotteException e) throws LinotteException {
		// Cas standard
		Ring codeProtege = ((JobContext) getContext()).getLivre().getKernelStack().fermeCodeProtege(((JobContext) getContext()).getLivre());
		if (codeProtege == null)
			throw e;
		CalqueParagraphe calqueParagraphe = ((JobContext) getContext()).getLivre().getKernelStack().recupereDernierCalqueParagraphe();
		calqueParagraphe.addActeurLocal(new Acteur(null, new Chaine("numéro_erreur"), Role.NOMBRE, new BigDecimal(e.getErreur()), null));
		String message = e.getToken();
		if (message == null)
			message = Messages.retourneErreur(String.valueOf(e.getErreur()));
		calqueParagraphe.addActeurLocal(new Acteur(null, new Chaine("message_erreur"), Role.TEXTE, message, null));
		currentProcessus = codeProtege.getProcessus();
		((JobContext) getContext()).getLivre().getKernelStack().recupereDerniereCondition().setEtat(ETAT_CONDITION.FAUX);
		return currentProcessus != null;
	}

	public Processus getCurrentProcessus() {
		return currentProcessus;
	}

	@Override
	public Object clone() {
		Job job = new Job((AKJobContext) ((JobContext) getContext()).clone());
		// job.pere = pere;
		// TODO, bug avec le webonotte !!
		job.pere = this;
		getSons().add(job);
		job.setFirstProcessus(getFirstProcessus());
		return job;
	}

	@Override
	public void stop() {
		setRunning(false);
		super.stop();
		if (isSon() && !pere.isDead()) {
			pere.stop();
		}
	}

	public boolean isSon() {
		return pere != null;
	}

	public void stopWithException(Exception e) {
		erreurAPropager = e;
		if (isSon()) {
			pere.stopWithException(e);
		}
		setRunning(false);
	}

	private void afficheDeboggueur() {
		RuntimeContext runtimeContext = (RuntimeContext) getRuntimeContext();
		JobContext jobContext = (JobContext) getContext();

		try {
			if (runtimeContext.getLibrairie().getToilePrincipale().isAtelier()) {
				FrameProcess.applet.getJButtonPause().setVisible(true);
				FrameProcess.applet.getJButtonContinuer().setVisible(true);
				FrameProcess.applet.getJButtonLire().setVisible(false);
				FrameProcess.applet.getInspecteur().setVisible(true);
				if (currentProcessus != null) {
					FrameProcess.applet.getInspecteur().getLabelDebogage().setText("Prochain état : " + currentProcessus.getEtat().clef());
				}
				if (jobContext.getLivre().getParagraphe() != null)
					FrameProcess.applet.getInspecteur().refresh(jobContext.getLivre().getParagraphe().getActeurs());
				FrameProcess.applet.setVisible(true);
				SwingUtilities.invokeLater(new Runnable() {
					public void run() {
						try {
							FrameProcess.applet.getCahierCourant().setPositionCahier(getCurrentProcessus().getPosition());
							FrameProcess.applet.getCahierCourant().getEditorPaneCahier().getLineHighlighter().setPositionDeboguage(getCurrentProcessus().getPosition());
						} catch (BadLocationException e) {
						}
					}
				});
				synchronized (this) {
					wait();
				}
			}
		} catch (InterruptedException e) {
			e.printStackTrace();
		} finally {
		}
	}

	public void setDebogueur(boolean debogueur) {
		this.debogueur = debogueur;
	}

	public Job getPere() {
		return pere;
	}

	@Override
	public String toString() {
		Livre livre = ((JobContext) getContext()).getLivre();
		if (livre != null) {
			String s = livre.getNom();
			if (livre.getParagraphe() != null) {
				s += " § " + livre.getParagraphe().toString();
			}
			return s;
		} else {
			return "sans nom";
		}
	}

	public String toShortString() {
		Livre livre = ((JobContext) getContext()).getLivre();
		if (livre != null) {
			String s = "";
			if (livre.getParagraphe() != null) {
				s += livre.getParagraphe().toString();
			}
			return s;
		} else {
			return "";
		}
	}

	public boolean isParallelise() {
		return parallelise;
	}

	public void setParallelise() {
		parallelise = true;
	}
}
