package org.linotte.moteur.xml.alize.kernel;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.EmptyStackException;
import java.util.Iterator;
import java.util.List;

import org.linotte.moteur.entites.Acteur;
import org.linotte.moteur.entites.Livre;
import org.linotte.moteur.outils.ArrayIterator;
import org.linotte.moteur.outils.Chaine;
import org.linotte.moteur.xml.appels.Appel;
import org.linotte.moteur.xml.appels.Appel.APPEL;
import org.linotte.moteur.xml.appels.Boucle;
import org.linotte.moteur.xml.appels.CalqueParagraphe;
import org.linotte.moteur.xml.appels.Condition;
import org.linotte.moteur.xml.appels.Fonction;
import org.linotte.moteur.xml.appels.Ring;
import org.linotte.moteur.xml.appels.SousParagraphe;
import org.linotte.moteur.xml.appels.TantQue;

@SuppressWarnings("serial")
public final class KernelStack implements Serializable, Cloneable {

	private Appel[] contents;

	private int size;

	private int initialSize;

	/**
	 * Creates a new empty stack.
	 */
	public KernelStack() {
		initialSize = 20;
		contents = new Appel[initialSize];
	}

	/**
	 * Returns <code>true</code> if the stack is empty, and <code>false</code>
	 * otherwise.
	 * 
	 * @return A boolean.
	 */
	boolean isEmpty() {
		return size == 0;
	}

	/**
	 * Returns the number of elements in the stack.
	 * 
	 * @return The element count.
	 */
	public int size() {
		return size;
	}

	/**
	 * Pushes an object onto the stack.
	 * 
	 * @param o
	 *          the object.
	 */
	public void push(Appel o) {
		final int oldSize = size;
		size += 1;
		if (contents.length == size) {
			// grow ..
			final Appel[] newContents = new Appel[size + initialSize];
			System.arraycopy(contents, 0, newContents, 0, size);
			contents = newContents;
		}
		contents[oldSize] = o;
	}

	/**
	 * Returns the object at the top of the stack without removing it.
	 * 
	 * @return The object at the top of the stack.
	 */
	public Appel peek() {
		return contents[size - 1];
	}

	/**
	 * Removes and returns the object from the top of the stack.
	 * 
	 * @return The object.
	 */
	public Appel pop() {
		if (size == 0) {
			throw new EmptyStackException();
		}
		size -= 1;
		final Appel retval = contents[size];
		contents[size] = null;
		return retval;
	}

	/**
	 * Returns a clone of the stack.
	 * 
	 * @return A clone.
	 */
	public Object clone() {
		try {
			KernelStack stack = (KernelStack) super.clone();
			if (contents != null) {
				contents = (Appel[]) contents.clone();
			}
			return stack;
		} catch (CloneNotSupportedException cne) {
			throw new IllegalStateException("Clone not supported? Why?");
		}
	}

	/**
	 * Clears the stack.
	 */
	public void clear() {
		size = 0;
		Arrays.fill(contents, null);
	}

	/**
	 * Returns the item at the specified slot in the stack.
	 * 
	 * @param index
	 *          the index.
	 * 
	 * @return The item.
	 */
	public Appel get(final int index) {
		return contents[index];
	}

	public Iterator<Appel> iterator() {
		Appel[] shrink = new Appel[size];
		System.arraycopy(contents, 0, shrink, 0, size);
		return new ArrayIterator<Appel>(shrink);
	}

	public void ajouterAppel(Appel o) {
		push(o);
	}

	public Appel retirerDernierAppel() {
		return pop();
	}

	public Appel regarderDernierAppel() {
		return peek();
	}

	public Condition fermeCondition() {
		return (Condition) close(APPEL.CONDITION);
	}

	public Boucle fermeBoucle() {
		return (Boucle) close(APPEL.BOUCLE);
	}

	public Fonction fermeFonction() {
		return (Fonction) close(APPEL.FONCTION);
	}

	public SousParagraphe fermeSousParagraphe() {
		return (SousParagraphe) close(APPEL.SOUSPARAGRAPHE);
	}

	public CalqueParagraphe fermeCalqueParagraphe() {
		return (CalqueParagraphe) close(APPEL.CALQUEPARAGRAPHE);
	}

	public Condition recupereDerniereCondition() {
		return (Condition) find(APPEL.CONDITION);
	}

	public Boucle recupereDerniereBoucle() {
		return (Boucle) find(APPEL.BOUCLE);
	}

	public List<Boucle> recupereToutesLesBoucles() {
		List<Boucle> list = new ArrayList<Boucle>();
		Appel retval;
		for (int s = size; s > 0;) {
			s -= 1;
			retval = contents[s];
			if (retval.getType() == APPEL.BOUCLE)
				list.add((Boucle) retval);
		}
		return list;
	}

	public Fonction recupereDerniereFonction() {
		return (Fonction) find(APPEL.FONCTION);
	}

	public SousParagraphe recupereDernierSousParagraphe() {
		return (SousParagraphe) find(APPEL.SOUSPARAGRAPHE);
	}

	public CalqueParagraphe recupereDernierCalqueParagraphe() {
		return (CalqueParagraphe) find(APPEL.CALQUEPARAGRAPHE);
	}

	public Acteur rechercheActeurDansSousParagraphe(Chaine n) {
		Appel appel;
		for (int s = size; s > 0;) {
			appel = contents[--s];
			if (appel.getType() == APPEL.FONCTION || appel.getType() == APPEL.CALQUEPARAGRAPHE)
				return null;
			else if (appel.getType() == APPEL.SOUSPARAGRAPHE) {
				SousParagraphe sousParagraphe = (SousParagraphe) appel;
				Acteur acteur = sousParagraphe.getActeurLocal(n);
				if (acteur != null)
					return acteur;
			}
		}
		return null;
	}

	public Acteur rechercheActeur(Chaine n) {
		Appel appel;
		for (int s = size; s > 0;) {
			appel = contents[--s];
			if (appel.getType() == APPEL.CALQUEPARAGRAPHE) {
				CalqueParagraphe paragraphe = (CalqueParagraphe) appel;
				Acteur acteur = paragraphe.getActeurLocal(n);
				if (acteur != null)
					return acteur;
			} else if (appel.getType() == APPEL.SOUSPARAGRAPHE) {
				SousParagraphe sousParagraphe = (SousParagraphe) appel;
				Acteur acteur = sousParagraphe.getActeurLocal(n);
				if (acteur != null)
					return acteur;
			} else if (appel.getType() == APPEL.FONCTION) {
				// http://langagelinotte.free.fr/forum/showthread.php?tid=996&pid=6838#pid6838
				return null;
			} 
		}
		return null;
	}

	/**
	 * @param retour
	 * @return
	 */
	public Acteur rechercheJokerDansBoucle(Acteur retour) {
		if (size() != 0) {
			// On recherche la dernière boucle :
			Boucle b = null;
			for (int i = size() - 1; i > -1; i--) {
				if (get(i) instanceof Boucle && !(get(i) instanceof TantQue)) {
					b = (Boucle) get(i);
					retour = b.retourJoker();
					break;
				}
			}
		}
		return retour;
	}

	public Ring fermeCodeProtege(Livre livre) {
		Appel retval;
		for (; size > 0;) {
			retval = contents[--size];
			contents[size] = null;
			if (retval.getType() == APPEL.FONCTION)
				// Bogue remonté par Pat :
				livre.setParagraphe(((Fonction) retval).getParagraphe());
			if (retval.getType() == APPEL.RING)
				return (Ring) retval;
		}
		return null;
	}

	private Appel close(APPEL c) {
		Appel retval;
		for (; size > 0;) {
			size -= 1;
			retval = contents[size];
			contents[size] = null;
			if (retval.getType() == c)
				return retval;
		}
		return null;
	}

	private Appel find(APPEL c) {
		Appel retval;
		for (int s = size; s > 0;) {
			s -= 1;
			retval = contents[s];
			if (retval.getType() == c)
				return retval;
		}
		return null;
	}

}
